import os
import pickle
import socket
import select
import signal
import string
import sys
import tempfile
import types
import whrandom

import einRpc
import niftySocket


version = (0, 4, 1)

# Aribtrary resources and limits.
DEFAULT_PORT = 2012
MAX_PORT = 2112
MAX_QUEUED_REQUESTS = 5

error = "BrendaError"


class TupleServer:
    """Uses fork for concurrency."""

    def __init__(self, host, detach=1):
	# It would be nice to streamline this beast so that it used
	# exec to spawn a new process without any of the baggage
	# currently in memory.  Could call this module as a script
	# with special args to do that.

	sock = niftySocket.FileSocket(socket.AF_INET, socket.SOCK_STREAM)

	success = 0
	for port in range(DEFAULT_PORT, MAX_PORT+1):
	    try:
		sock.bind(host, port)
	    except socket.error:
		pass
	    else:
	        success = 1
	        break
	    # end try
	# end for

	if not success:
	    raise error, \
	        "Could not start server on any port within limits"
	# end if

	sock.listen(MAX_QUEUED_REQUESTS)  # First listen

	self._port = port
	self._serverUp = 1
	self._tuples = []  # List of tuples
	self._clientConns = []  # List of sockets (as files)

	# These map the client's socket (file) to its tuple fragment.
	self._pendingIn = {}
	self._pendingRd = {}

	# Fork the server off, but return here so the caller can get
        # the port info.
	if detach:
	    pid = os.fork()

	    if (pid == 0):
	        self.__serverLoop(sock)
		sys.exit(0)
	    else:
		self._pid = pid
		#sock.close()
	    # end if
	else:
	    print ":ts", self, "started on port", self._port
	    self.__serverLoop(sock)
	# end if
    # end def __init__

    def __serverLoop(self, listenSock):
	while self._serverUp:
	    readyList = select.select([listenSock] + self._clientConns,
				      [], [])
	    for sock in readyList[0]:
		# sock is either listenSock or a client connection
		if (sock is listenSock):
		    conn, addr = sock.accept()
		    print ":ts got conn", conn, "from", addr
		    self._clientConns.append(conn)
		else:
		    #print ":ts got request from", sock
		    self._dispatchRequest(sock)
		# end if
	    # end for

	# end while
    # end def __serverLoop

    def _dispatchRequest(self, file):
	# Data format:
	# <command>\n<...>
	# <command> is one of "in", "out", or "rd".
	# <...> are the arguments to the command, if any.  These are
	# command specific, but are usually a string or pickled tuple.

	command = file.readline()

	if (command == ''):
	    # EOF reached; close the socket.
	    self._closeClient(file)
	    return
	# end if

	command = string.strip(command)
	#print ":ts command is", command
	# Dispatch on the command.
	if (command == 'in'):
	    t = self._readPickledOrError(file, types.TupleType)
	    print ":ts:", file, "pend (in)", repr(t)
	    self._matchAndWrite(file, t, 0)
	elif (command == 'rd'):
	    t = self._readPickledOrError(file, types.TupleType)
	    print ":ts:", file, "pend (rd)", repr(t)
	    self._matchAndWrite(file, t, 1)
	elif (command == 'out'):
	    t = self._readPickledOrError(file, types.TupleType)
	    print ":ts:", file, "->", repr(t)
	    self._tuples.append(t)
	    self._fulfillPendingReqs(len(self._tuples) - 1)
	else:
	    self._writeErrorResponse(file, 'unk command "' + command + '"')
	# end if
    # end def _dispatchRequest

    def _closeClient(self, file):
	i = self._clientConns.index(file)
	file.close()
	print ":ts:", file, "closed"
	del self._clientConns[i]
    # end def _closeClient

    def _readPickledOrError(self, file, desiredType):
	#print ":ts getting data"

	x = pickle.load(file)
	if (type(x) != desiredType):
	    # Send an error response back to client
	    self._writeErrorResponse(file, "type mismatch")
	# end if
	#print ":ts got data, type is correct"
	return x
    # end def _readPickledOrError

    def _writeErrorResponse(self, file, msg):
	fullMessage = ("error\n" + msg + "\n")
	file.write(fullMessage)
	print ":ts:", file, "error:", msg
    # end def _writeErrorResponse

    def _fulfillPendingReqs(self, i):
	# Match the list of pending "in" and "rd" requests against the
	# tuple at index i.
	#print ":ts fulfillPendingReqs()"
	t = self._tuples[i]
	result = None
	for clientFile in self._pendingIn.keys():
	    if self._matchTuple(self._pendingIn[clientFile], t):
		self._sendTupleAt(i, clientFile, 0)
		del self._pendingIn[clientFile]
		return
	    # end if
	# end for

	for clientFile in self._pendingRd.keys():
	    if self._matchTuple(self._pendingRd[clientFile], t):
		self._sendTupleAt(i, clientFile, 1)
		del self._pendingRd[clientFile]
	    # end if
	# end for

    # end def _fulfillPendingReqs

    def _matchAndWrite(self, file, frag, rd):
	#print ":ts matchAndWrite()"
	i = self._getMatchingIndex(frag, self._tuples)
	if i < 0:
	    #print ":ts not matched"
	    assert not self._hasPending(file), "multiple reqs from client?!"
	    if rd:
		self._pendingRd[file]=frag
	    else:
		self._pendingIn[file]=frag
	    # end if
	    #print ":ts added to pending"
	else:
	    #print ":ts matched"
	    self._sendTupleAt(i, file, rd)
	# end if
    # end def

    def _sendTupleAt(self, i, file, rd):
	"""Fill a tuple request.

	Arguments:
	    i    - the index into self._tuples; this is the tuple to send
	    file - the destination client socket (as a file object)
	    rd   - flag; if false, the tuple is removed from self._tuples
	Return None.

	"""

	t = self._tuples[i]

	file.write("ok\n" + pickle.dumps(t))

	if not rd:
	    del self._tuples[i]
	# end if
	print ":ts:", file, "<-", repr(t)
    # end def _sendTupleAt

    def _getMatchingIndex(self, frag, ls):
	for i in range(len(ls)):
	    t = ls[i]
	    if self._matchTuple(frag, t):
		return i
	    # end if
	# end for
	return -1
    # end def

    def _hasPending(self, file):
	return (file in self._pendingRd.keys() or 
	        file in self._pendingIn.keys())
    # end def

    def _matchTuple(self, frag, t):
	"""return yes/no match"""

	result=1
	if (len(t) < len(frag)):
	    return 0
        # end if
	for i in range(len(frag)):
	    if frag[i] != t[i]:
		result = 0
		break
	    # end if
	# end for

	return result
    # end def _matchTuple

# end class TupleServer



class TupleClientCds:

    """Code development system interface to PyBrenda.

    Do nothing in parallel; simulate using multiple processes on this
    machine.  Also serve as a base class.

    XXX how to use.

    """

    def __init__(self, host=None, port=None):
	"""Create a connection to the tuple server.

	Arguments:
	host - string	Default:  <localhost>
	port - int      Default:  None

	Specifying None for the port means to start a new server on
	host.

	"""

	if host is None:
	    host = socket.gethostname()
	# end if

	if port is None:
	    self._server = TupleServer(host)
	    port = self._server._port
	else:
	    self._server = None
	# end if

	self._host = host
	self._port = port
	self._conn = self._connectTo(host, port)
    # end def __init__

    def __del__(self):
        """Destructor.  Clean up if I started the server."""

	if self._server:
	    pid = self._server._pid
	    os.kill(pid, signal.SIGTERM)
	# end if
    # end def __del__

    def _connectTo(self, host, port):
	sock = niftySocket.FileSocket(socket.AF_INET,
				      socket.SOCK_STREAM)
	sock.connect(host, port)
	return sock
    # end def _connectTo

    def Eval(self, preTuple, func, args, postTuple=()):
	# Check types
	if (type(preTuple) != types.TupleType):
	    raise ValueError, ("not a tuple: " + repr(preTuple))
	elif (type(func) != types.BuiltinFunctionType and
	      type(func) != types.FunctionType):
	    raise ValueError, ("not a function:  " + repr(func))
	elif (type(args) != types.TupleType):
	    raise ValueError, ("not a tuple: " + repr(args))
	elif (type(postTuple) != types.TupleType):
	    raise ValueError, ("not a tuple: " + repr(postTuple))
	# end if

	if os.fork() == 0:
	    c = TupleClientCds(self._host, self._port)
	    workerReturned = apply(func, (c,)+args)
	    #print ":tc: preTuple is", repr(preTuple)
	    #print ":tc: worker returned", repr(workerReturned)
	    #print ":tc: postTuple is", repr(postTuple)
	    apply(c.Out, (preTuple + (workerReturned,) + postTuple))
	    sys.exit(0)
	# end if
    # end def Eval

    def In(self, *frag):
	return self._requestTuple(frag, 0)
    # end def In

    def Rd(self, *frag):
	return self._requestTuple(frag, 1)
    # end def Rd

    def _requestTuple(self, frag, rd):
	if rd:
	    self._conn.write("rd\n")
	else:
	    self._conn.write("in\n")
	# end if

	pickle.dump(frag, self._conn)

	respHdr = string.strip(self._conn.readline())

	result = None
	if (respHdr=="ok"):
		result = pickle.load(self._conn)
	else:
		msg = self._conn.readline()
	# end if

	if result is not None:
		return result
	else:
		raise error, ("got bad response from server: " + 
			      respHdr + ": " + msg)
	# end if
    # end def _requestTuple

    def Out(self, *t):
	self._conn.write("out\n")
	pickle.dump(t, self._conn)
    # end def Out

# end class TupleClientCds


class TupleClientRsh(TupleClientCds):
    """Uses rsh for parallel Eval()."""

    def __init__(self, host=None, port=None):
	TupleClientCds.__init__(self, host, port)

	# Where to find the resource file and create the temp dir.
	# Path should be accessible to all hosts (e.g., via NFS).
	self._topDir = os.environ['HOME']

	# Open the .PyBrenda.rc file to find out what machines I
	# should use.
	f = open(os.path.join(self._topDir, '.PyBrenda.rc'),
	         'r')
	targetStr = 'PyBrenda.Appl.nodelist: '
	while 1:
	    s = f.readline()
	    if string.find(s, targetStr) == 0:
	        nodelist = s[len(targetStr):-1]  # chop newline too
		self._nodelist = string.split(nodelist)
		break
	    elif not s:
		raise error, 'No nodelist specified in .PyBrenda.rc'
		break
	    # end if
	# end while

	# Choose a starting host at random.
	self._nodeIndex = whrandom.randint(0, len(self._nodelist)-1)

	# Create area for temporary scripts.
	self._tmpDir = os.path.join(self._topDir, '.PyBrenda.tmp')
	try:
	    os.mkdir(self._tmpDir)
	except:
	    pass
	# end try
    # end def __init__

    def Eval(self, preTuple, func, args, postTuple=()):
	# Check types
	if (type(preTuple) != types.TupleType):
	    raise ValueError, ("not a tuple: " + repr(preTuple))
	elif (type(func) != types.BuiltinFunctionType and
	      type(func) != types.FunctionType):
	    raise ValueError, ("not a function:  " + repr(func))
	elif (type(args) != types.TupleType):
	    raise ValueError, ("not a tuple: " + repr(args))
	elif (type(postTuple) != types.TupleType):
	    raise ValueError, ("not a tuple: " + repr(postTuple))
	# end if

	# Create the RPC object.
	req = einRpc.SimpleRpcRequest(func, args)

	# Temporarily use my tmpDir, then restore the tempfile module.
	old_tempdir = tempfile.tempdir
	tempfile.tempdir = self._tmpDir
	tmpFile = tempfile.mktemp()
	tempfile.tempdir = old_tempdir

	# Create temporary script file.
	scp = open(tmpFile, 'w')
	scp.write('import sys\n')
	scp.write('sys.path = ' + repr(sys.path) + '\n')
	scp.write('\n')
	scp.write('import os\n')
	scp.write('os.chdir(' + repr(os.getcwd()) + ')\n')
	scp.write('\n')
	scp.write('import tempfile\n')
	scp.write('tempfile.template = None\n')
	scp.write('\n')
	scp.write('import pickle\n')
	scp.write('import Brenda\n')
	scp.write('\n')
	scp.write('req = pickle.loads(' + repr(pickle.dumps(req)) + ')\n')
	scp.write('c = Brenda.TupleClientRsh(' + repr(self._host) + 
	          ', ' + repr(self._port) + ')\n')
	scp.write('workerReturned = req.invoke((c,))\n')
	scp.write('apply(c.Out, (' + repr(preTuple) + 
		  '+(workerReturned,)+' + repr(postTuple) +'))\n')
	scp.close()

	if (os.fork() == 0):
	    command = ('rsh ' + self._nodelist[self._nodeIndex] + 
		       ' python ' + tmpFile) 
	    print ":tcr: command is", command
	    os.system(command)
	    #os.unlink(tmpFile)
	    sys.exit(0)
	# end if

	self._nodeIndex = (self._nodeIndex + 1) % len(self._nodelist)
    # end def Eval

# end class TupleClientRsh
